package framework.device.graphics
{
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.display.Graphics;
	import flash.display.Shader;
	import flash.display.Shape;
	import flash.geom.ColorTransform;
	import flash.geom.Matrix;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import framework.Maths;
	import framework.resource.ResourceImage;
	import framework.util.Rect;
	import framework.util.RectEx;
	import framework.util.Vec2D;
	
	/**
	 * 描画管理クラス
	 * @author syun
	 */
	public class GraphicManager
	{
		/**
		 * スクリーン
		 */
		private var m_Graphics:Graphics;
		private var m_Screen:BitmapData;
		private var m_Capture:Bitmap; // キャプチャー画面
		private var m_Rect:Rect;
		private var m_Vec:Vec2D;
		private var m_Shake:Vec2D;
		private var m_FadeTimer:int; // フェード用タイマー
		private var m_FadeTimerMax:int; // フェードタイマーの最大値
		private var m_FadeColor:int; // フラッシュの色
		
		/**
		 * 描画色
		 */
		private var m_Color:uint = 0x000000;
		/**
		 * 半透明
		 */
		private var m_Alpha:Number = 1.0;
		/**
		 * コンストラクタ
		 * @param	screen
		 */
		public function GraphicManager(screen:BitmapData, g:Graphics) 
		{
			m_Graphics = g;
			ResourceImage.setValue("__default", new Bitmap(screen));
			m_Screen = screen;
			m_Capture = new Bitmap(new BitmapData(screen.width, screen.height, false));
			ResourceImage.setValue("__capture", m_Capture);
			m_Rect = new Rect();
			m_Vec = new Vec2D();
			m_Shake = new Vec2D();
			m_FadeTimer = 0;
			m_FadeTimerMax = 0;
			m_FadeColor = 0x000000;
		}
		
		/**
		 * レンダリングスクリーンを変更する
		 * @param	key キー。何も指定しないと、デフォルトスクリーン
		 */
		public function changeScreen(key:String="__default"):void
		{
			m_Screen = ResourceImage.getValue(key).bitmapData;
		}
		
		/**
		 * フェードを開始する
		 * @param	timer タイマー
		 * @param	color 色
		 */
		public function beginFade(timer:int, color:int):void
		{
			m_FadeTimerMax = timer;
			m_FadeTimer = timer;
			m_FadeColor = color;
		}
		
		/**
		 * フェードの描画
		 */
		public function drawFade():void
		{
			if (m_FadeTimer > 0)
			{
				var a:Number = m_FadeTimer / m_FadeTimerMax;
				setColor(m_FadeColor, a);
				fillRect();
				m_FadeTimer--;
			}
		}
		
		/**
		 * ビットマップバッファを作る
		 * @param	key キー
		 * @param	width 幅
		 * @param	height 高さ
		 * @param	transparent ピクセル単位の透過色を有効にするかどうか
		 * @param	fillColor 塗りつぶし色
		 */
		public function createBuffer(key:String, width:int, height:int, transparent:Boolean=true, fillColor:uint=0xFFFFFFFF):void
		{
			var bitmap:Bitmap = new Bitmap(new BitmapData(width, height, transparent, fillColor));
			ResourceImage.setValue(key, bitmap);
		}
		
		/**
		 * 画面を揺らす
		 * @param	x
		 * @param	y
		 */
		public function shake(x:Number, y:Number):void
		{
			m_Shake.set(x, y);
		}
		public function shakeBegin():void
		{
			m_Screen.scroll(m_Shake.x, m_Shake.y);
		}
		public function shakeEnd():void
		{
			m_Screen.scroll(0, 0);
		}
		
		public function getCapture():Bitmap
		{
			return m_Capture;
		}
		/**
		 * キャプチャーする
		 */
		public function capture():void
		{
			m_Capture.bitmapData.draw(m_Screen);
		}
		
		/**
		 * キャプチャー画面の描画（拡大α）
		 * @param	v 0～1のパラメータ
		 */
		public function drawCaptureScale(v:Number):void
		{
			var ct:ColorTransform = new ColorTransform(1, 1, 1, v);
			var m:Matrix = new Matrix();
			var s:Number = 1 + (1 - v) * 5;
			m.translate(-m_Capture.width / 2, -m_Capture.height / 2);
			m.scale(s, s);
			m.translate(m_Capture.width / 2, m_Capture.height / 2);
			m_Screen.draw(m_Capture, m, ct);
		}
		/**
		 * キャプチャー画面の描画（分割して描画）
		 * @param	v 0～1のパラメータ
		 */
		public function drawCaptureDiv(v:Number):void
		{
			var dSize:int = 64; // 分割サイズ
			var ofs:Number = dSize * (1-v) / 2;
			for (var j:int = 0; j < m_Capture.bitmapData.height/dSize; j++)
			{
				for (var i:int = 0; i < m_Capture.bitmapData.width/dSize; i++)
				{
					var x:Number = i * dSize + ofs;
					var y:Number = j * dSize + ofs;
					draw2(m_Capture, x, y, x, y, dSize*v, dSize*v);
				}
			}
		}
		
		/**
		 * キャプチャー画面の描画（横方向に引き伸ばす）
		 * @param	v 0～1のパラメータ
		 */
		public function drawCaptureStrechV(v:Number):void
		{
			var sx:Number = (2 - v) * (2 - v) * (2 - v);
			var sy:Number = v * v;
			draw2(m_Capture, 0, 0, 0, 0, 0, 0, sx, sy);
		}
		
		/**
		 * 半透明の設定
		 * @param	alpha
		 */
		public function setAlpha(alpha:Number):void
		{
			m_Alpha = alpha;
		}
		/**
		 * 描画色の設定
		 * @param	color 色
		 * @param alpha 半透明値
		 */
		public function setColor(color:uint = 0x000000, alpha:Number=1):void
		{
			m_Color = color;
			m_Alpha = alpha;
		}
		
		/**
		 * 塗りつぶしなしの矩形描画
		 * @param	x
		 * @param	y
		 * @param	w
		 * @param	h
		 */
		public function drawRect(x:Number, y:Number, w:Number, h:Number):void
		{
			var x1:Number = x;
			var y1:Number = y;
			var x2:Number = x + w;
			var y2:Number = y + h;
			drawLine(x1, y1, x2, y1);
			drawLine(x2, y1, x2, y2);
			drawLine(x1, y2, x2, y2);
			drawLine(x1, y1, x1, y2);
		}
		/**
		 * 高速な矩形描画
		 * @param	rect
		 */
		public function fillRect2(rect:Rectangle):void
		{
			m_Screen.fillRect(rect, m_Color);
		}
		public function fillRectEx2(rect:RectEx):void
		{
			fillRect2(rect.getRectangle());
		}
		
		/**
		 * 矩形の描画（塗りつぶし）
		 * @param	x
		 * @param	y
		 * @param	w
		 * @param	h
		 * @param	rot
		 */
		public function fillRect(x:Number=0, y:Number=0, w:Number=-1, h:Number=-1, rot:Number=0):void
		{
			if (w == -1 || h == -1)
			{
				w = Main.WIDTH;
				h = Main.HEIGHT;
			}
			var s:Shape = new Shape();
			var m:Matrix = new Matrix();
			s.graphics.beginFill(m_Color, m_Alpha);
			s.graphics.drawRect( -w / 2, -h / 2, w, h);
			s.graphics.endFill();
			m.rotate(Maths.deg2rad(rot));
			m.translate(x+w/2, y+h/2);
			m_Screen.draw(s, m);
		}
		/**
		 * 矩形の描画（中心座標を指定）
		 * @param	cx 中心座標X
		 * @param	cy 中心座標Y
		 * @param	w 幅
		 * @param	h 高さ
		 * @param rot 回転角度
		 */
		public function fillRectEx(cx:Number, cy:Number, w:Number, h:Number, rot:Number=0):void
		{
			var x:Number = cx - w;
			var y:Number = cy - h;
			fillRect(x, y, w * 2, h * 2, rot);
		}
		
		/**
		 * 線を引く
		 * @param	x1
		 * @param	y1
		 * @param	x2
		 * @param	y2
		 */
		public function drawLine(x1:Number, y1:Number, x2:Number, y2:Number):void
		{
			var rect:Rectangle;
			if (x1 == x2)
			{
				rect = new Rectangle(x1, y1, 1, y2 > y1 ? y2-y1 : y1 - y2);
				fillRect2(rect);
				return;
			}
			if(y1 == y2)
			{
				rect = new Rectangle(x1, y1, x2 > x1 ? x2 - x1 : x1 - x2, 1);
				fillRect2(rect);
				return;
			}
			var s:Shape = new Shape();
			s.graphics.lineStyle(1, m_Color, m_Alpha);
			s.graphics.moveTo(x1, y1);
			s.graphics.lineTo(x2, y2);
			m_Screen.draw(s);
		}
		
		public function drawTriangleEx(cx:Number, cy:Number, w:Number, rot:Number):void
		{
			var vertices:Vector.<Number> = new Vector.<Number>;
			vertices.push(0); // x1
			vertices.push(-w); // y1
			vertices.push(-w * 0.866); // x2 = -w / √3 / 2
			vertices.push(+w * 0.5);
			vertices.push(+w * 0.866);
			vertices.push(+w * 0.5);
			
			var s:Shape = new Shape();
			var m:Matrix = new Matrix();
			m.rotate(Maths.deg2rad(rot));
			m.translate(cx, cy);
			
			s.graphics.beginFill(m_Color, m_Alpha);
			s.graphics.drawTriangles(vertices);
			s.graphics.endFill();
			m_Screen.draw(s, m);
		}
		
		public function drawTriangle(x1:Number, y1:Number, x2:Number, y2:Number, x3:Number, y3:Number, rot:Number):void
		{
			var minX:Number = Math.min(x1, x2, x3);
			var minY:Number = Math.min(y1, y2, y3);
			var maxX:Number = Math.max(x1, x2, x3);
			var maxY:Number = Math.max(y1, y2, y3);
			
			// ３点の中心にする場合はこっち
			//var cx:Number = (minX + maxX) / 2; // 中心座標(X)
			//var cy:Number = (minY + maxY) / 2; // 中心座標(Y)
			
			// 重心で回転させる場合はこっち
			var cx:Number = (x1 + x2 + x3) / 3; // 中心座標(X)
			var cy:Number = (y1 + y2 + y3) / 3; // 中心座標(Y)
			
			var vertices:Vector.<Number> = new Vector.<Number>;
			vertices.push(x1 - cx);
			vertices.push(y1 - cy);
			vertices.push(x2 - cx);
			vertices.push(y2 - cy);
			vertices.push(x3 - cx);
			vertices.push(y3 - cy);
			
			var s:Shape = new Shape();
			var m:Matrix = new Matrix();
			m.rotate(Maths.deg2rad(rot));
			m.translate(cx, cy);
			
			s.graphics.beginFill(m_Color, m_Alpha);
			s.graphics.drawTriangles(vertices);
			s.graphics.endFill();
			m_Screen.draw(s, m);
		}
		
		/**
		 * 三角形の描画
		 * @param	vertices 頂点リスト
		 * @param	indices 頂点インデックスリスト
		 */
		public function drawTriangles(vertices:Vector.<Number>, indices:Vector.<int>=null):void
		{
			var s:Shape = new Shape();
			s.graphics.beginFill(m_Color, m_Alpha);
			s.graphics.drawTriangles(vertices, indices);
			s.graphics.endFill();
			m_Screen.draw(s);
		}
		
		/**
		 * 円の描画 (塗りつぶしなし)
		 * @param	cx 中心座標(X)
		 * @param	cy 中心座標(Y)
		 * @param	r 半径
		 * @param	thick 線の太さ
		 */
		public function drawCircle(cx:Number, cy:Number, r:Number, thick:Number=1):void
		{
			var s:Shape = new Shape();
			s.graphics.lineStyle(thick, m_Color, m_Alpha);
			s.graphics.drawCircle(cx, cy, r);
			m_Screen.draw(s);
		}
		
		/**
		 * 円の描画 (塗りつぶす)
		 * @param	cx 中心座標(X)
		 * @param	cy 中心座標(Y)
		 * @param	r 半径
		 */
		public function fillCircle(cx:Number, cy:Number, r:Number):void
		{
			var s:Shape = new Shape();
			s.graphics.beginFill(m_Color, m_Alpha);
			s.graphics.drawCircle(cx, cy, r);
			s.graphics.endFill();
			m_Screen.draw(s);
		}
		
		/**
		 * 画像をスクロールする
		 * @param	key
		 * @param	dx
		 * @param	dy
		 */
		public function scrollImage(key:String, dx:Number, dy:Number):void
		{
			//var img:Bitmap = getImage(key);
			//img.scrollRect = new Rectangle(dx, dy, img.width-dx, img.height-dy);
		}
		
		/**
		 * 画像情報を取得する
		 * @param	key
		 * @return
		 */
		public function getImage(key:String):Bitmap
		{
			return ResourceImage.getValue(key);
		}
		
		/**
		 * 画像の幅を取得する
		 * @param	key
		 * @return
		 */
		public function getWidth(key:String):int
		{
			var img:Bitmap = getImage(key);
			return img.width;
		}
		
		/**
		 * 画像の高さを取得する
		 * @param	key
		 * @return
		 */
		public function getHeight(key:String):int
		{
			var img:Bitmap = getImage(key);
			return img.height;
		}
		
		/**
		 * 画像の描画
		 */
		public function draw(key:String, x:Number, y:Number, rot:Number=0, sx:Number=1, sy:Number=1):void
		{
			var img:Bitmap = getImage(key);
			var m:Matrix = new Matrix();
			if (rot != 0)
			{
				// 中心を原点に持っていってから回転させる
				var w:Number = img.width/2;
				var h:Number = img.height/2;
				m.translate(-w, -h);
				m.rotate(Maths.deg2rad(-rot));
				m.translate(w, h);
			}
			m.translate(x, y);
			m_Screen.draw(img, m, null, img.blendMode);
		}
		public function drawEx(key:String, cx:Number, cy:Number, rot:Number=0, sx:Number=1, sy:Number=1):void
		{
			var img:Bitmap = getImage(key);
			var x:Number = cx - img.width / 2;
			var y:Number = cy - img.height / 2;
			draw(key, x, y, rot);
		}
		
		public function draw2(img:Bitmap, x:Number, y:Number, ox:Number=0, oy:Number=0, ow:Number=0, oh:Number=0, sx:Number=1, sy:Number=1):void
		{
			var m:Matrix = new Matrix();
			if (ow == 0 || oh == 0)
			{
				ow = img.width;
				oh = img.height
			}
			img.scrollRect = new Rectangle(ox, oy, ow, oh);
			var ow2:Number = ow / 2;
			var oh2:Number = oh / 2;
			m.translate(-ow2, -oh2);
			m.scale(sx, sy);
			m.translate(x+ow2, y+oh2);
			m_Screen.draw(img, m);
		}
		public function draw2Ex(img:Bitmap, cx:Number, cy:Number, ox:Number=0, oy:Number=0, ow:Number=0, oh:Number=0, sx:Number=1, sy:Number=1):void
		{
			var x:Number = cx - img.width / 2;
			var y:Number = cy - img.height / 2;
			draw2(img, x, y, ox, oy, ow, oh, sx, sy);
		}
		
		/**
		 * 高速なBitmap描画
		 * @param	key
		 * @param	x
		 * @param	y
		 * @param	ox
		 * @param	oy
		 * @param	ow
		 * @param	oh
		 */
		public function drawFast(key:String, x:Number, y:Number, ox:Number=0, oy:Number=0, ow:Number=0, oh:Number=0):void
		{
			var img:Bitmap = getImage(key);
			drawFast2(img, x, y, ox, oy, ow, oh);
		}
		
		public function drawFast2(img:Bitmap, x:Number, y:Number, ox:Number = 0, oy:Number = 0, ow:Number = 0, oh:Number = 0):void
		{
			if (ow == 0 || oh == 0)
			{
				ow = img.width;
				oh = img.height;
			}
			
			m_Rect.set(ox, oy, ow, oh);
			m_Vec.set(x, y);
			m_Screen.copyPixels(img.bitmapData, m_Rect, m_Vec.getPoint());
		}
		
		public function drawFastEx(key:String, cx:Number, cy:Number, ox:Number = 0, oy:Number = 0, ow:Number = 0, oh:Number = 0):void
		{
			if (ow == 0 || oh == 0)
			{
				var img:Bitmap = getImage(key);
				ow = img.width;
				oh = img.height;
			}
			var x:Number = cx - ow / 2;
			var y:Number = cy - oh / 2;
			drawFast(key, x, y, ox, oy, ow, oh);
		}
		
		/**
		 * 回転の連番画像を描画する
		 * @param	key
		 * @param	pattern
		 * @param	cx
		 * @param	cy
		 * @param	deg
		 */
		public function drawRot(key:String, pattern:int, cx:Number, cy:Number, deg:Number):void
		{
			var img:Bitmap = getImage(key);
			var rot:int = Math.floor(deg);
			if (rot < 0) { rot = 360 + rot; }
			var idx:Number = Math.floor(rot / (360 / pattern));
			var size:Number = img.width / pattern;
			var ox:Number = idx * size;
			drawFastEx(key, cx, cy, ox, 0, size, size);
		}
		
		/**
		 * フォントの描画
		 * @param	f
		 */
		public function drawFont(f:FontObj):void
		{
			m_Screen.draw(f.getTextField(), f.getMatrix());
		}
	}

}